////////////////////////////////////
// clock AND protoThreads configure
#include "config_1_3_2.h" 

// threading library w/ python interface
#include "pt_cornell_1_3_2_python.h" 

////////////////////////////////////
// graphics libraries
// for debugging only
#include "tft_master.h"
#include "tft_gfx.h"
#include <math.h>
#include "i2c_helper.h"

// The fixed point types===========NOT CHECKED (NC)=============
#include <stdfix.h>
#include <stdlib.h>
// end of NC

// thread structures 
//static struct pt pt_timer;
static struct pt pt_print;
static struct pt pt_imu, pt_led;
static struct pt pt_buttons, pt_sliders, pt_radio, pt_serial;

// VARIABLES ======================================================
// Debugging with TFT
//char buffer[60];   // sys_time_seconds printf

// Python Interface =========================
static float og_vel = 20.0;
static char r = 255; // WHITE
static char g = 255; // WHITE
static char b = 255; // WHITE 
static int mode = 1; // 1 = realistic ride, 2 = drop tower, 3 = bouncer

static int length = 5; // number of LEDs that constitute the cart
static int reset = 0;  // soft reset; to 

//static int const_start_length_rc = 10; // number of leds for constant velocity at the beginning
                               // of the roller coaster 
                               // range: 0 - 20 
//static float const_start_vel_rc = 5; // not absolute velocity; controlled by yield 
                               // range: 1 - 10 
static float vel_ratio = 0.04; // includes friction 
                               // range: 0 - 1 with interval 0.01
static float acc_ratio = 0.9;  // control sensitivity to the accelerometer readings 
                               // range: 0 - 1 with interval 0.01
// end of Python Interface====================

// IMU =======================================
#define NUM_IMU 7 // number of accelerometers
//#define COORDINATES 3       // acceleration coordinates: x, y, z
#define IMU_READ_PERIOD 50 // in ms
static float imu_readings[NUM_IMU];
unsigned int imu_ado[NUM_IMU];
// end of IMU ================================

// Dotstar ===================================
// APA102 datasheet:
// 32 bits of zeros is a start frame
// 32 bits of ones is a stop frame
// LED frame:
// 111_5bit_global_intensity_8bitBlue_8bitGreen_8bitRed
// so 0xff_00_ff_00 is full intnsity green
#define START_FRAME 0x00000000
#define STOP_FRAME 0xffffffff
#define PIXEL_FRAME(i, r, g, b) (0xe0000000 | (((0x1f & (i))) << 24) | ((0xff & (b)) << 16) | ((0xff & (g)) << 8) | (0xff & (r)))
//#define PIXEL_FRAME(i,r,g,b)(0xe0000000 | ((i)<<24) | ((b)<<16) | ((g)<<8) | (r))
#define FULL_ON 0x1e
#define HALF_ON 0x0f
#define QUAR_ON 0x07
#define LED_OFF 0x00

// number of pixels
#define PixelNum 120
// number of accelerometers
//#define AccelNum NUM_IMU

// time interval in sec
#define TIMEINT 1.0
#define LEDINT (PixelNum / (NUM_IMU- 1))
#define LED_UPDATE_INTERVAL 5 // number of thread yielding in msec 
#define LED_TIME_INTERVAL 25   // seconds in the led world proportional to real world 

// Should be in real RC metrics ('real' m and sec; conversion happens automatically
// in the code)
// Real roller coaster specs: 100 m in length, top speed 120 km/h = 33 m/s,
// LED roller coaster specs: 2 m in length, 120 leds, 1.67 cm/led = 0.000167 m/led
#define RATIO_M 0.02 // Real RC : LED RC = 50 : 1 
#define RATIO_T 0.04 // Real time : LED time = 100 : 1
#define RATIO_LED 0.0167 // 2m : 120 LED = 1 : 60 

static int done = 0;
static int rand_wait = 2048; // in msec, for drop tower
static int rising = 1;  // drop tower 
// static int const_start = 1; // roller coaster 

typedef struct pixel pixel;
struct pixel
{
  char red;
  char green;
  char blue;
  char intensity;
};
pixel pixel_array[PixelNum];
// end of Dotstar =============================

// Positional variables =======================
// - in real world metrics: TIMEINT, curr_vel 
// - in led RC metrics : init_y, init_vel, curr_acc
static float init_vel = 20.0;
static float curr_vel = 2; 
static float curr_acc = 0; 
static float tmp_y = 0;
static int init_y = 0; 
static int curr_y = 5;
// end of Positional variables =================

// system 1 second interval tick
static int sys_time_seconds;

// === display the LEDs ===========================================
// copies the contents of pixel_array to SPI
void write_pixels(void)
{
  // start frame
  WriteSPI2(START_FRAME);
  // wait for end of transaction
  while (SPI2STATbits.SPIBUSY)
    ;

  int i;
  //payload
  for (i = 0; i < PixelNum; i++)
  {
    WriteSPI2(PIXEL_FRAME(pixel_array[i].intensity, pixel_array[i].red, pixel_array[i].green, pixel_array[i].blue));
    // wait for end of transaction
    while (SPI2STATbits.SPIBUSY)
      ;
  }
  //stop frame
  WriteSPI2(STOP_FRAME);
  // wait for end of transaction
  while (SPI2STATbits.SPIBUSY)
    ;
}

// === write a RGBI value to the pixel array =======================
void set_pixel_rgb(int i, char r, char g, char b, char intensity)
{
  if (i < 0 || i >= PixelNum)
    return;
  pixel_array[i].intensity = intensity; //enforce max
  pixel_array[i].red = r;
  pixel_array[i].green = g;
  pixel_array[i].blue = b;
}

// === LED Thread =================================================
static PT_THREAD(protothread_led(struct pt *pt))
{
  PT_BEGIN(pt);
  
  // tmp variables
  //  static int i = 0; // should be counting up like sys_time_seconds
  static int i, j;        // counting variable
  // Initialize to OFF
  for (i = 0; i < PixelNum; i++)
  {
    set_pixel_rgb(i, 0, 0, 0, LED_OFF);
  }
  write_pixels();

  while (1)
  {
    //
//    PT_YIELD_TIME_msec(LED_UPDATE_INTERVAL);

    // RESET
    if (reset == 1)
    {
      curr_y = 0;
      curr_vel = og_vel;
//      const_start = 1;
      tmp_y = 0;
      rising = 1;
      done = 0;
      
      reset = 0;
    }

    // acceleration conversion
    // imu_reading ranges -16000 ~ 16000 in adc
    // raw imu data in m/s^2
    // (imu data) * ratio_m / ratio_t^2
    if (curr_y <= LEDINT/2)
    {
      curr_acc = imu_readings[0];
    }
    else if (curr_y > LEDINT/2 && curr_y <= LEDINT * 3/2)
    {
      curr_acc = imu_readings[1];
    }
    else if (curr_y > LEDINT *3/2 && curr_y <= LEDINT * 5/2)
    {
      curr_acc = imu_readings[2];
    }
    else if (curr_y > LEDINT *5/2 && curr_y <= LEDINT * 7/2)
    {
      curr_acc = imu_readings[3];
    }
    else if (curr_y > LEDINT *7/2 && curr_y <= LEDINT * 9/2)
    {
      curr_acc = imu_readings[4];
    }
    else if (curr_y > LEDINT *9/2 && curr_y <= LEDINT * 11/2)
    {
      curr_acc = imu_readings[5];
    } 
    else
    {
      curr_acc = imu_readings[6];
    }
    // @ HERE: 
    // - in real world metrics: TIMEINT, curr_vel 
    // - in led RC metrics : init_y, init_vel, curr_acc
    // TIMEINT ignored because it's just 1
    // 
    //    #define RATIO_M 0.02 // Real RC : LED RC = 50 : 1 
    //    #define RATIO_T 0.04 // Real time : LED time = 100 : 1

//    const_start_vel_rc = 5;
    
    if (mode == 1) { // mode 1 : real time roller coaster
//        if (const_start) {
//            for (i = 0; i < const_start_length_rc; i++){ // constant speed at the beginning 
//                for (j = 0; j < PixelNum; j++)
//                    {
//                      if (j == i)
//                      {
//                        set_pixel_rgb(j, r, g, b, QUAR_ON);
//                      } 
//                      else 
//                      {
//                        set_pixel_rgb(j, r, g, b, LED_OFF);
//                      }
//                    }
//                write_pixels();
//                PT_YIELD_TIME_msec(5 * const_start_vel_rc);
//            }
//            const_start = 0;
//            tmp_y = const_start_length_rc;
//        }
//        else {
        PT_YIELD_TIME_msec(LED_UPDATE_INTERVAL);

        if (done == 0) {
            init_vel = curr_vel; // velocity conversion 
            tmp_y = (float)tmp_y + init_vel * vel_ratio + 0.5 * curr_acc * acc_ratio; // acceleration conversion_ratio; // acceleration conversion
            curr_y = (int)(tmp_y * RATIO_LED); // float to int 
            // Update init_vel and init_y
            curr_vel = curr_acc + curr_vel; // time interval = 1000 msec
        }
        if (curr_y > PixelNum){ // ride is finished
            done = 1;
            curr_y = PixelNum - 1;
        }
//        }
    } else if (mode == 2) { // drop tower
        if (rising) { 
            for (i = PixelNum - 1; i >= length; i--){
                for (j = 0; j < PixelNum; j++) {
                    if (j <= i && j>i-length) {
                        set_pixel_rgb(j, r, g, b, QUAR_ON);
                    } 
                    else {
                        set_pixel_rgb(j, r, g, b, LED_OFF);
                    } 
                }
                write_pixels();
                PT_YIELD_TIME_msec(30);
            }
            
            rand_wait = ((rand() & 0xffff) >> 5) + (1 << 11); // 2048 to 4096
            PT_YIELD_TIME_msec(3000); // wait a random amount of time to increase the SUSPENSE
            rising = 0;
            tmp_y = length;
        } else {
            PT_YIELD_TIME_msec(10); 
            if (done == 0) {
                init_vel = curr_vel; 
                tmp_y = (float)tmp_y + init_vel + 0.5 * curr_acc; 
                curr_y = (int)(tmp_y * RATIO_LED); // float to int
                // Update init_vel and init_y
                curr_vel = curr_acc + curr_vel; // time interval = 1000 msec
            }
            if (curr_y > PixelNum){ // ride is finished
                done = 1;
                curr_y = PixelNum - 1;
            }
        }
    } else if(mode==3) { // the bouncer
        PT_YIELD_TIME_msec(LED_UPDATE_INTERVAL);
        if(curr_y < PixelNum){
            init_vel = curr_vel; // velocity conversion 
            tmp_y = (float)tmp_y + init_vel * vel_ratio + 0.5 * curr_acc * acc_ratio; // acceleration conversion_ratio; // acceleration conversion
            //tmp_y = (float)tmp_y + init_vel + 0.5 * curr_acc;
            curr_y = (int)(tmp_y * RATIO_LED); // float to int 
            // Update init_vel and init_y
            curr_vel = curr_acc + curr_vel; // time interval = 1000 msec
        } else {
            init_vel = curr_vel; // velocity conversion 
            tmp_y = (float)tmp_y - init_vel * vel_ratio + 0.5 * curr_acc * acc_ratio; // acceleration conversion_ratio; // acceleration conversion
            //tmp_y = (float)tmp_y - init_vel + 0.5 * curr_acc;
            curr_y = (int)(tmp_y * RATIO_LED); // float to int 
            // Update init_vel and init_y
            curr_vel = curr_acc - curr_vel; // time interval = 1000 msec
        }      
    }
    
    for (j = 0; j < PixelNum; j++) {
        if (curr_y>=j && curr_y < j+length ) {
            set_pixel_rgb(j, r, g, b, QUAR_ON);
        } 
        else {
            set_pixel_rgb(j, r, g, b, LED_OFF);
        }
    }
    // Update LED
    write_pixels();
  }
  PT_END(pt);
} // led thread

// === Print Thread ===============================================
// for debugging only
static PT_THREAD(protothread_print(struct pt *pt))
{
  PT_BEGIN(pt);
  static int q;
  while (1)
  {
    PT_YIELD_TIME_msec(1000);

    sys_time_seconds++;   // updated in led thread
    printf("System time : %d sec \r", sys_time_seconds);
    for (q = 0; q < NUM_IMU; q++){
        printf("imu %i: %f \r", q, imu_readings[q]);
    }
    printf("curr_acc: %f \r", curr_acc);
    printf("curr_vel: %f \r", curr_vel);
    printf("init_vel: %f \r", init_vel);
    printf("og_vel: %f \r", og_vel);
    printf("curr_y: %i \r", curr_y);
    printf("mode: %i \r", mode);
    // NEVER exit while
  } // END WHILE(1)
  PT_END(pt);
} // animation thread

static int c, d;
static int values[3];
// === IMU thread =================================================
static PT_THREAD(protothread_imu(struct pt *pt))
{
  PT_BEGIN(pt);
    
    while(1) {
        for (c = 0; c < NUM_IMU; c++){
            // slave address to read from is 0xd0
            // set all other addresses to 0xd2
            mPORTBClearBits(imu_ado[c]);
            
            // Read data from IMU in format
            // {xAccel, yAccel, zAccel}
            readImuValues(values); // returns results in ADC units

            // Parse the IMU data
            // 16 bit resolution, +- 2^15 = 32768; range +- 2g
            // note: can't shift to divide because resolution gets lost
//            imu_readings[c] = values[1];
            values[1] = (values[1] >> 4) << 4;
            imu_readings[c] = (values[1] * 9.8) / pow(2, 14) * -1;
            
            for (d = 0; d < NUM_IMU; d++){
                mPORTBSetBits(imu_ado[d]);
            }
        }
        PT_YIELD_TIME_msec(IMU_READ_PERIOD);
    }
  PT_END(pt);
}

// === outputs from python handler =============================================
// signals from the python handler thread to other threads
// These will be used with the prototreads PT_YIELD_UNTIL(pt, condition);
// to act as semaphores to the processing threads
char new_string = 0;
char new_button = 0;
char new_toggle = 0;
char new_slider = 0;
char new_list = 0;
char new_radio = 0;
// identifiers and values of controls
// current button
char button_id, button_value;
// current toggle switch/ check box
char toggle_id, toggle_value;
// current radio-group ID and button-member ID
char radio_group_id, radio_member_id;
// current slider
int slider_id;
float slider_value; // value could be large
// current listbox
int list_id, list_value;
// current string
char receive_string[64];

// === Buttons thread ==========================================================
// process buttons from Python
static PT_THREAD (protothread_buttons(struct pt *pt))
{
    PT_BEGIN(pt);
    // set up LED port A0 to blink
    while(1){
        PT_YIELD_UNTIL(pt, new_button==1);
        // clear flag
        new_button = 0;   
        if (button_id==1 && button_value==1){ // pushed
            reset = 1;
//            printf("reset \r");
        }
    } // END WHILE(1)   
    PT_END(pt);  
} // thread blink

// ===  Slider thread =========================================================
// process sliders from Python
static PT_THREAD(protothread_sliders(struct pt *pt))
{
  PT_BEGIN(pt);
  while (1)
  {
    PT_YIELD_UNTIL(pt, new_slider == 1);
    // clear flag
    new_slider = 0;
    if (slider_id == 1)
    {
      curr_vel = (float)slider_value;
      og_vel = (float)slider_value;
    }
    if (slider_id == 2)
    {
      vel_ratio = (float)slider_value;
    }
    if (slider_id == 3)
    {
      acc_ratio = (float)slider_value;
    }
    if (slider_id == 4)
    {
      r = (char)slider_value;
    }
    if (slider_id == 5)
    {
      g = (char)slider_value;
    }
    if (slider_id == 6)
    {
      b = (char)slider_value;
    }
    if (slider_id == 7)
    {
      length = (int)slider_value;
    }
  } // END WHILE(1)
  PT_END(pt);
} // thread slider

// ===  radio thread =========================================================
// process radio buttons from Python
static PT_THREAD (protothread_radio(struct pt *pt))
{
    PT_BEGIN(pt);
    while(1){
        PT_YIELD_UNTIL(pt, new_radio==1);
        // clear flag
        new_radio = 0; 
        if (radio_group_id == 1){
            if (radio_member_id == 1){
                mode = 1 ;
            }
            if (radio_member_id == 2){
                mode = 2 ;
            }
            if (radio_member_id == 3){
                mode = 3 ;
            }
            reset = 1; 
        }
    } // END WHILE(1)   
    PT_END(pt);  
} // thread radio

// === Python serial thread ====================================================
static PT_THREAD(protothread_serial(struct pt *pt))
{
  PT_BEGIN(pt);
  static char junk;
  
  while (1)
  {
    // There is no YIELD in this loop because there are
    // YIELDS in the spawned threads that determine the
    // execution rate while WAITING for machine input
    // =============================================
    // NOTE!! -- to use serial spawned functions
    // you MUST edit config_1_3_2 to
    // (1) uncomment the line -- #define use_uart_serial
    // (2) SET the baud rate to match the PC terminal
    // =============================================

    // now wait for machine input from python
    // Terminate on the usual <enter key>
    PT_terminate_char = '\r';
    PT_terminate_count = 0;
    PT_terminate_time = 0;
    // note that there will NO visual feedback using the following function
    PT_SPAWN(pt, &pt_input, PT_GetMachineBuffer(&pt_input));

    // Parse the string from Python
    // There can be toggle switch, button, slider, and string events

    // toggle switch
    if (PT_term_buffer[0] == 't')
    {
      // signal the button thread
      new_toggle = 1;
      // subtracting '0' converts ascii to binary for 1 character
      toggle_id = (PT_term_buffer[1] - '0') * 10 + (PT_term_buffer[2] - '0');
      toggle_value = PT_term_buffer[3] - '0';
    }

    // pushbutton
    if (PT_term_buffer[0] == 'b')
    {
      // signal the button thread
      new_button = 1;
      // subtracting '0' converts ascii to binary for 1 character
      button_id = (PT_term_buffer[1] - '0') * 10 + (PT_term_buffer[2] - '0');
      button_value = PT_term_buffer[3] - '0';
    }

    // slider
    if (PT_term_buffer[0] == 's')
    {
      sscanf(PT_term_buffer, "%c %d %f", &junk, &slider_id, &slider_value);
      new_slider = 1;
    }

    // listbox
    if (PT_term_buffer[0] == 'l')
    {
      new_list = 1;
      list_id = PT_term_buffer[2] - '0';
      list_value = PT_term_buffer[3] - '0';
      //printf("%d %d", list_id, list_value);
    }

    // radio group
    if (PT_term_buffer[0] == 'r')
    {
      new_radio = 1;
      radio_group_id = PT_term_buffer[2] - '0';
      radio_member_id = PT_term_buffer[3] - '0';
      //printf("%d %d", radio_group_id, radio_member_id);
    }

    // string from python input line
    if (PT_term_buffer[0] == '$')
    {
      // signal parsing thread
      new_string = 1;
      // output to thread which parses the string
      // while striping off the '$'
      strcpy(receive_string, PT_term_buffer + 1);
    }
  } // END WHILE(1)
  PT_END(pt);
} // thread blink

void main(void)
{
  // === config threads ==========
  // turns OFF UART support and debugger pin, unless defines are set
  PT_setup();
  // === setup system wide interrupts  ========
  INTEnableSystemMultiVectoredInt();
  
  srand(1); // for the drop tower random wait
  
  // init the threads
  PT_INIT(&pt_print);
  PT_INIT(&pt_imu);
  PT_INIT(&pt_led);
  PT_INIT(&pt_sliders);
  PT_INIT(&pt_serial);
  PT_INIT(&pt_radio);
  PT_INIT(&pt_buttons);
  
  // TFT SETUP ================================================================
  // init the display
  // NOTE that this init assumes SPI channel 1 connections
  // for debugging only
//    tft_init_hw();
//    tft_begin();
//    tft_fillScreen(ILI9340_BLACK);
//    tft_setRotation(0); // 240x320 vertical display; use tft_setRotation(1) for 320x240
  // ==========================================================================

  // ACCELEROMETER SETUP ========================================================
  // Enable I2C channel
  OpenI2C1(I2C_ON, 0xC2);

  // all on channel B
  imu_ado[0] = BIT_0; 
  imu_ado[1] = BIT_1;
  imu_ado[2] = BIT_2;
  imu_ado[3] = BIT_3;
  imu_ado[4] = BIT_7;
  imu_ado[5] = BIT_11;
  imu_ado[6] = BIT_13;

  // Take the MPU 6050 out of sleep mode and set accelerometer range to +-2g
  char data[] = {0};
  
  int x;
  for (x = 0; x < NUM_IMU; x++){
    mPORTBSetPinsDigitalOut(imu_ado[x]);
    mPORTBClearBits(imu_ado[x]);
    i2c_write(0x6b, data, 1);
    i2c_write(0x1c, data, 1);
    mPORTBSetBits(imu_ado[x]);
  }

  // DOTSTAR SETUP ============================================================
  // divide Fpb by 2, configure the I/O ports. Not using SS 
  // 16 bit transfer CKP=1 CKE=1
  // possibles SPI_OPEN_CKP_HIGH;   SPI_OPEN_SMP_END;  SPI_OPEN_CKE_REV
  // For any given peripheral, you will need to match these
  SpiChnOpen(2, SPI_OPEN_ON | SPI_OPEN_MODE32 | SPI_OPEN_MSTEN | SPICON_CKP, 4);
  // SCK2 is pin 26
  // SDO2 (MOSI) is in PPS output group 2, could be connected to RB5 which is pin 14
  PPSOutput(2, RPB5, SDO2);

  int i;
  // Initialize LED strip
  for (i = 0; i < PixelNum; i++)
  {
    set_pixel_rgb(i, 0, 0, 0, LED_OFF);
  }
  write_pixels();

  //===========================================================================

  while (1) {
//      PT_SCHEDULE(protothread_print(&pt_print)); // for debugging only
      PT_SCHEDULE(protothread_imu(&pt_imu));
      PT_SCHEDULE(protothread_led(&pt_led));
      PT_SCHEDULE(protothread_buttons(&pt_buttons));
      PT_SCHEDULE(protothread_sliders(&pt_sliders));
      PT_SCHEDULE(protothread_radio(&pt_radio));
      PT_SCHEDULE(protothread_serial(&pt_serial));
  }
}